#!/usr/bin/env ruby
#
# scroball.rb USER BACKDATE PATH...
#
# Add all .mp3 files found (recursively) under each PATH to the recent
# history in last.fm for user USER.  Note that USER's password will be
# read interactively from standard input.  The entries are backdated
# so that the first one starts at BACKDATE.  BACKDATE is seconds since
# the UNIX epoch; `/bin/date -d "HUMAN DATE" +%s` may be of use
# generating it, if you have GNU coreutils.
#
# Copyright 2007, 2008 Dean Scarff <dos@scarff.id.au>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License.  You
# may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

begin
require 'rubygems'
rescue LoadError
end
require 'net/http'
require 'mp3info'
require 'digest/md5'
require 'uri'
require 'iconv' # until String#toutf8 works

MAX_BULK_SUBMISSION = 50
user = ARGV.shift
BACKDATE = Time.at(ARGV.shift.to_i)
print 'Password: '
`stty -echo`
pass = $stdin.gets.chomp
`stty echo`
print "\n"
t = Time.new.to_i
if 0 == t
  error "invalid time"
end

def md5(x)
  Digest::MD5.hexdigest(x)
end
def urlenc(x)
  begin
    x = Iconv.iconv("utf-8", "iso-8859-1", x)[0]
  rescue Iconv::IllegalSequence, Iconv::InvalidCharacter
  end
  x.gsub(/[^a-zA-Z0-9~_\.\-]/n) {|s| sprintf('%%%02x', s[0]) }
end

## HANDSHAKE ###########################################################

hpath = sprintf('/?hs=true&p=1.2&c=%s&v=%s&u=%s&t=%s&a=%s',
                'sba', '0.1', user, t, md5(md5(pass) + t.to_s))
warn hpath
hres = Net::HTTP.get_response('post.audioscrobbler.com', hpath)
hres.value # just for the error
if !hres.body.nil?
  hlines = hres.body.split($/)
  hstatus = hlines[0]
  if 'OK' != hstatus
    raise "Handshake response: " + hstatus
  end
else
  raise "no body in handshake response"
end

@sessionid = hlines[1]
nowplaying = hlines[2]
@submission = URI.parse(hlines[3])
warn hres.body

## PARSING AND SUBMISSION ##############################################

@i = 0
@t = BACKDATE.utc.to_f # this is the *start* time
@postdata = []

def formpair(a, k, v)
  a << urlenc(k) + '=' + urlenc(v)
end

def submit
  sreq = Net::HTTP.new(@submission.host, @submission.port)
  sres = sreq.request_post(@submission.path,
                           @postdata.unshift(["s=" + @sessionid]).join('&'))
  warn sres.body
  @postdata.clear
  @i = 0
end

def mp3_entry(path)
  x = []
  Mp3Info.open(path) do |mp3|
    (mp3.tag.artist && mp3.tag.title && mp3.length) or
      raise "missing metadata"
    formpair(x, "a[#{@i}]", mp3.tag.artist)
    formpair(x, "t[#{@i}]", mp3.tag.title)
    formpair(x, "i[#{@i}]", (@t - mp3.length).to_i.to_s)
    formpair(x, "o[#{@i}]", 'P')
    formpair(x, "l[#{@i}]", mp3.length.to_i.to_s)
    formpair(x, "r[#{@i}]", '')
    formpair(x, "b[#{@i}]", mp3.tag.album)
    if mp3.tag.tracknum && 0 < mp3.tag.tracknum
      formpair(x, "n[#{@i}]", mp3.tag.tracknum.to_s)
    else
      formpair(x, "n[#{@i}]", '')
    end
    formpair(x, "m[#{@i}]", '')
    @t += mp3.length.ceil
  end
  x
end

def necessary?
  MAX_BULK_SUBMISSION == @i
end

ARGV.each do |arg|
  if (FileTest.file?(arg) && arg =~ /\.mp3$/)
    print "#{arg}\n"
    begin
      @postdata << mp3_entry(arg)
      @i += 1
    rescue
        warn "Problem with #{arg}: " + $!
    end
  elsif (FileTest.directory?(arg))
    Dir["#{arg}/**/*.mp3"].each do |path|
      next unless FileTest.readable?(path)
      print "#{path}\n"
      begin
        @postdata << mp3_entry(path)
        @i += 1
      rescue
        warn "Problem with #{path}: " + $!
      end
      submit if necessary?
    end
  end
  submit if necessary?
end

submit if @i > 0

print "Listening finished at: " +
  Time.at(@t.to_i).getlocal.strftime("%c %Z") + "\n"
